# -*- coding: utf-8 -*-
"""
Created on Tue Apr 26 12:36:57 2016

@author: kswider
"""
   
# Do podstawowych struktur danych Pythona należą słowniki, ktore kojarzą
# wartości z kluczami i pozwalają na szybki dostęp do wartości odpowiadającej
# zadanemu kluczowi.
   
pusty_slownik = {}                      # w stylu Pythona
pusty_slownik2 = dict()                 # mniej w stylu Pythona
punkty = { "Jan" : 80, "Anna" : 95 }    # literał słownikowy

# Dostęp do wartości dla pewnego klucza jest możliwy przy użyciu nawiasow
# kwadratowych:

punkty_Jana = punkty["Jan"]             # przyjmuje wartość 80

# W przypadku podania klucza, ktorego nie ma w słowniku pojawia się 
# sygnalizacja błędu KeyError.

try:
    punkty_Kasi = punkty["Kasia"]
except KeyError:
    print "nie ma punktow dla Kasi!"

#  Przy pomocy operatora in można sprawdzić, czy klucz występuje w słowniku:

Jan_ma_punkty = "Jan" in punkty         # True
Kasia_ma_punkty = "Kasia" in punkty     # False

# Słowniki mają metodę get, ktora w przypadku odwołania do nieistniejącego
# klucza zwraca wartość domyślną (zamiast generować wyjątek).

punkty_Jana = punkty.get("Jan", 0)      # przyjmuje wartość 80
punkty_Kasi = punkty.get("Kasia", 0)    # przyjmuje wartość 0
punkty_niczyje = punkty.get("Nikt")     # przyjmuje wrtość pustą

# Pary klucz-wartość przypisujemy używając nawiasow kwadratowych

punkty["Anna"] = 99         # zamienia poprzednią wartość
punkty["Kasia"] = 100       # dodaje trzecią pozycję
liczba_stud = len(punkty)   # wynosi 3

# Typowym użyciem słownikow jest przedstawianie danych strukturalnych

tweet = {
"uzytkownik" : "krzysztofswider",
"tekst" : "Analiza danych jest niesamowita",
"liczba_retweet" : 100,
"hashtags" : ["#dane", "#analiza", "#analizadanych", "#niesamowita", "#yolo"]
}

# Niezależnie od poszukiwania poszczeglnych kluczy można spojrzeć na nie
# łącznie:

tweet_klucze = tweet.keys()             # lista of kluczy
tweet_wartosci = tweet.values()         # lista wartości
tweet_pozycje = tweet.items()           # lista par (klucz, wartość)
"uzytkownik" in tweet_klucze            # True, but uses a slow list in
"uzytkownik" in tweet                   # bardziej w stylu Pythona
"krzysztofswider" in tweet_wartosci     # True

# Klucze słownika muszą być trwałe, w szczeglności nie mogą być listami.
# Jeśli chcemy aby klucz był wieloczęściowy, wowczas należy użyć krotki lub
# znaleźć sposob na zamiany klucza na łańcuch

# defaultdict
# Przypuśćmy, ża mamy policzyć słowa w dokumencie. Naturalnym sposobem jest
# utworzenie słownika w ktorym klucze są słowami, a wartości odpowiadają 
# liczbie słow. Z chwilą gdy napotykamy kolejne słowo, zwiększamy liczbę jego 
# wystąpień, o ile już jest w słowniku, bądź dodajemy do słownika, jeśli to
# nowe słowo.

liczniki_slow = {}
for slowo in dokument:
    if slowo in liczniki_slow:
        liczniki_slow[slowo] += 1
    else:
        liczniki_slow[slowo] = 1
        
# Można także postąpić według zasady "Lepiej wybaczyć niż zezwolić" i po prostu
# obsłużyć wyjątek wynikający z sytuacji, gdy poszukiwany jest nieistniejący
# klucz:
        
liczniki_slow = {}
for slowo in dokument:
    try:
        liczniki_slow[slowo] += 1
    except KeyError:
        liczniki_slow[slowo] = 1

# Można także użyć metody get ktora w elegancki sposob obsługuje brakujące
# klucze

liczniki_slow = {}
for slowo in dokument:
    poprzednia_liczba = liczniki_slow.get(slowo, 0)
    liczniki_slow[slowo] = poprzednia_liczba + 1
    
# Każdy z przytoczonych sposobow jest trochę nieporęczny, a można to poprawić
# poprzez użycie słownika defaultdict. Ma on cechy normalnego słownika, z tym,
# że w razie poszukiwania nieistniejącego klucza, jest dodawana
# dla niego wartość generowana przez wskazaną funkcję nie mającą argumentow.
# Użycie słownikow defaultdict wymaga zaimportowania go z modułu collections.
    
from collections import defaultdict

liczniki_slow = defaultdict(int) # int() zwraca 0
for slowo in dokument:
    liczniki_slow[slowo] += 1
    
# Słowniki domyślne defauultdict mogą także być przydatne w połączeniu z 
# funkcjami list lub dict, a nawet z funkcjami zdefiniowanymi przez
# programistę.

dd_listy = defaultdict(list)    # list() zwraca pustą listę
dd_listy[2].append(1)           # teraz dd_listy zawiera {2: [1]}

dd_slowniki = defaultdict(dict) # dict() zwraca pusty słownik
dd_slowniki["Jan"]["Miasto"] = "Lublin" # { "Jan" : { "Miasto" : Lublin"}}

dd_para = defaultdict(lambda: [0, 0])
dd_para[2][1] = 1               # teraz dd_para zawiera {2: [0,1]}

# Będą one przydatne w sytuacjach, kiedy używamy słownikow do "zbierania"
# wynikow identyfikowanych pewnym kluczem i nie chcemy za każdym razem 
# sprawdzać, czy klucz już istnieje.

# Counter
# Funkcja Counter zamienia ciąg wartości na obiekt w rodzaju 
# defaultdict(int) odwzorowując klucze na liczniki. Będzie ona przede 
# wszystkim wykorzystana do tworzenia histogramow.

from collections import Counter
c = Counter([0, 1, 2, 0])  # c ma (zasadniczo) postać { 0 : 2, 1 : 1, 2 : 1 }

# Pozwala to w bardzo prosty sposob rozwiązać problem zliczania słow:

liczniki_slow = Counter(dokument)

# Obiekt Counter ma często przydatną metodę most_common.

# Drukuj 10 najczęściej występujących słow wraz z liczbą ich wystąpień
for slowo, liczba_wyst in liczniki_slow.most_common(1):
    print slowo, liczba_wyst
    
# Jeśli pracujemy z łąńcuchem znakw, tzn. 

dokument1 = "Jan idzie i milczy"

# to poleceniem

liczniki_znakow = Counter(dokument1)

# zostaną zliczone wystąpienia poszczeglnych znakow w łańcuchu